@@ -10,7 +10,7 @@
 
 import re
 
-from mercurial import util, node, hg
+from mercurial import util, node, hg, error
 from mercurial.i18n import _
 
 import lfutil
@@ -26,10 +26,11 @@ class StoreError(Exception):
 
     def longmessage(self):
         return (_("error getting id %s from url %s for file %s: %s\n") %
-                 (self.hash, self.url, self.filename, self.detail))
+                 (self.hash, util.hidepassword(self.url), self.filename,
+                  self.detail))
 
     def __str__(self):
-        return "%s: %s" % (self.url, self.detail)
+        return "%s: %s" % (util.hidepassword(self.url), self.detail)
 
 class basestore(object):
     def __init__(self, ui, repo, url):
@@ -59,48 +60,58 @@ class basestore(object):
         missing = []
         ui = self.ui
 
-        util.makedirs(lfutil.storepath(self.repo, ''))
-
         at = 0
         available = self.exists(set(hash for (_filename, hash) in files))
         for filename, hash in files:
-            ui.progress(_('getting largefiles'), at, unit='lfile',
+            ui.progress(_('getting largefiles'), at, unit=_('files'),
                 total=len(files))
             at += 1
             ui.note(_('getting %s:%s\n') % (filename, hash))
 
             if not available.get(hash):
                 ui.warn(_('%s: largefile %s not available from %s\n')
-                        % (filename, hash, self.url))
+                        % (filename, hash, util.hidepassword(self.url)))
                 missing.append(filename)
                 continue
 
-            storefilename = lfutil.storepath(self.repo, hash)
-            tmpfile = util.atomictempfile(storefilename + '.tmp',
-                                          createmode=self.repo.store.createmode)
-
-            try:
-                hhash = self._getfile(tmpfile, filename, hash)
-            except StoreError, err:
-                ui.warn(err.longmessage())
-                hhash = ""
-            tmpfile.close()
-
-            if hhash != hash:
-                if hhash != "":
-                    ui.warn(_('%s: data corruption (expected %s, got %s)\n')
-                            % (filename, hash, hhash))
-                util.unlink(storefilename + '.tmp')
+            if self._gethash(filename, hash):
+                success.append((filename, hash))
+            else:
                 missing.append(filename)
-                continue
-
-            util.rename(storefilename + '.tmp', storefilename)
-            lfutil.linktousercache(self.repo, hash)
-            success.append((filename, hhash))
 
         ui.progress(_('getting largefiles'), None)
         return (success, missing)
 
+    def _gethash(self, filename, hash):
+        """Get file with the provided hash and store it in the local repo's
+        store and in the usercache.
+        filename is for informational messages only.
+        """
+        util.makedirs(lfutil.storepath(self.repo, ''))
+        storefilename = lfutil.storepath(self.repo, hash)
+
+        tmpname = storefilename + '.tmp'
+        tmpfile = util.atomictempfile(tmpname,
+                                      createmode=self.repo.store.createmode)
+
+        try:
+            gothash = self._getfile(tmpfile, filename, hash)
+        except StoreError as err:
+            self.ui.warn(err.longmessage())
+            gothash = ""
+        tmpfile.close()
+
+        if gothash != hash:
+            if gothash != "":
+                self.ui.warn(_('%s: data corruption (expected %s, got %s)\n')
+                             % (filename, hash, gothash))
+            util.unlink(tmpname)
+            return False
+
+        util.rename(tmpname, storefilename)
+        lfutil.linktousercache(self.repo, hash)
+        return True
+
     def verify(self, revs, contents=False):
         '''Verify the existence (and, optionally, contents) of every big
         file revision referenced by every changeset in revs.
@@ -170,8 +181,10 @@ def _openstore(repo, remote=None, put=False):
         lfpullsource = getattr(repo, 'lfpullsource', None)
         if lfpullsource:
             path = ui.expandpath(lfpullsource)
-        else:
+        elif put:
             path = ui.expandpath('default-push', 'default')
+        else:
+            path = ui.expandpath('default')
 
         # ui.expandpath() leaves 'default-push' and 'default' alone if
         # they cannot be expanded: fallback to the empty string,
@@ -196,7 +209,7 @@ def _openstore(repo, remote=None, put=False):
     try:
         storeproviders = _storeprovider[scheme]
     except KeyError:
-        raise util.Abort(_('unsupported URL scheme %r') % scheme)
+        raise error.Abort(_('unsupported URL scheme %r') % scheme)
 
     for classobj in storeproviders:
         try:
@@ -204,4 +217,5 @@ def _openstore(repo, remote=None, put=False):
         except lfutil.storeprotonotcapable:
             pass
 
-    raise util.Abort(_('%s does not appear to be a largefile store') % path)
+    raise error.Abort(_('%s does not appear to be a largefile store') %
+                     util.hidepassword(path))
