@@ -1,5 +1,14 @@
-import os, sys, time
-from mercurial import hg, ui, commands, util
+from __future__ import absolute_import
+
+import os
+import sys
+import time
+from mercurial import (
+    commands,
+    hg,
+    ui as uimod,
+    util,
+)
 
 TESTDIR = os.environ["TESTDIR"]
 BUNDLEPATH = os.path.join(TESTDIR, 'bundles', 'test-no-symlinks.hg')
@@ -8,8 +17,11 @@ BUNDLEPATH = os.path.join(TESTDIR, 'bundles', 'test-no-symlinks.hg')
 if not getattr(os, "symlink", False):
     sys.exit(80) # SKIPPED_STATUS defined in run-tests.py
 
+u = uimod.ui()
+# hide outer repo
+hg.peer(u, {}, '.', create=True)
+
 # clone with symlink support
-u = ui.ui()
 hg.clone(u, {}, BUNDLEPATH, 'test0')
 
 repo = hg.repository(u, 'test0')
@@ -21,7 +33,7 @@ commands.status(u, repo)
 # now disable symlink support -- this is what os.symlink would do on a
 # non-symlink file system
 def symlink_failure(src, dst):
-    raise OSError, (1, "Operation not permitted")
+    raise OSError(1, "Operation not permitted")
 os.symlink = symlink_failure
 
 # dereference links as if a Samba server has exported this to a
@@ -33,10 +45,10 @@ for f in 'test0/a.lnk', 'test0/d/b.lnk':
     fp.close()
 
 # reload repository
-u = ui.ui()
+u = uimod.ui()
 repo = hg.repository(u, 'test0')
 commands.status(u, repo)
 
 # try cloning a repo which contains symlinks
-u = ui.ui()
+u = uimod.ui()
 hg.clone(u, {}, BUNDLEPATH, 'test1')
